package no.rwr.x;
import gnu.gcj.RawData;


// Lightweight Visual 

/* 

   The visual defines the pixel encoding/decoding:
                                                
   Static Color		-		HWLUT   predefined
   Pseudo Color		-		HWLUT	colormap 
   Static Gray          -       	        1:1             gray 
   Gray Scale		-               HWLUT   colormap        gray
   True Color		Decomposed RGB          1:1
   Direct Color		Decomposed RGB  HWLUTs	colormap


   gray-value ---> <gray?> [yes] -----------.
                   [no] ---> v, v, v ---.   |
                                        :   |
   RGB ---> <gray?> [yes] ---> toGray(RGB) -`-> v ---> <HWLUT?>
	    [no]                        :          _ [yes]  [no] _
             |                          |         /    |     |    \
             !--------------------------'        |     |  (scale)  |
             |                                   |  iLUT[v]  |     |
            <RGB-decomposition?>                  \___ | ___ | ___/ 
            [yes]  [no] ---> findPixel(RGB) -----------`-----.    \
             |                                               |     \
             |                                               |     may be
            <HWLUTs?>                                        |    combined
            [yes] ---> riLUT[R], giLUT[G], biLUT[B] --.      |   to a single
            [no]                                      |      |  precalculated
             |	                                      |      |     lookup
   ,---------'----------------------------------------'      |
   |                                                         |
   `---> r, g, b ---> composeRGB(r, g, b) ---------------> pixel
                                                             |
   ,-------------- ...to server... <-------------------------'   
   |                                                          
   `---> drawable ---> HW ---> <RGB-decomposition?> [no] --.          
                               [yes] ---> R-v, G-v, B-v    |
                                           |    |    |     |
            ,------------------------------'----'----'-----'
            |                            (in parallel)
         <HWLUT?> [no] ----.
	 [yes] ---> LUT ---`---> value


   True color:

       RGB ---> composeRGB(r, g, b) ---> pixel

       gray-value ---> composeRGB(v, v, v) ---> pixel

       // composeRGB must use the bitmasks of the visual
       // Optimization: Hardcode for certain bitmasks


   Pseudo color/Static color:

       RGB ---> findPixel(RGB) ---> pixel

       gray-value ---> findPixel(vvv) ---> pixel

       // Optimization: findPixel() can be a simple lookup. This is easy
       // to do for gray-values.


   Gray scale/Static gray:

       RGB ---> lookup[toGray(RGB)] ---> pixel

       gray-value ---> lookup[v] ---> pixel

       // Optimization: lookup can be avoided if gray scale and
       // using same scale.
       // toGray() can be tuned to return the correct scale.
       // I.e.
       // Gray scale:
       //     RGB ---> toGray(RGB) ---> pixel
       //     gray-value ---> pixel


   Direct color:

       RGB ---> composeRGB(riLUT[R], giLUT[G], biLUT[B]) ---> pixel

       grayValue ---> composeRGB(riLUT[v], giLUT[v], biLUT[v]) ---> pixel

       // Optimization: Create iLUTs in preshifted state.
       // Composing is simply ORing the lookups.

*/

/** 
    A visual determines how a color is encoded into a pixel/bitfield.
    It does not determine how the pixel/bitfield is encoded into an image.
    
    <p>This class encapsulates all three Xlib representations of a visual.

    <ul>

    <li>int: visual id.

    <li>Visual: opaque data structure used by a lot of Xlib functions.

    <li>VisualInfo: transparent data structure that binds the visual
    to a certain screen and depth.

    </ul>

    <p>This class does not examine nor manipulate the Visual
    structure, since the X manual says the structure is opaque, and
    that XVisualInfo should be used instead.


 */
public final class Visual {

    public static final int VC_STATIC_GRAY  = 0;
    public static final int VC_GRAY_SCALE   = 1;
    public static final int VC_STATIC_COLOR = 2;
    public static final int VC_PSEUDO_COLOR = 3;
    public static final int VC_TRUE_COLOR   = 4;
    public static final int VC_DIRECT_COLOR = 5;

    protected static final int MASK_ID            = 1 << 0;
    protected static final int MASK_SCREEN        = 1 << 1;
    protected static final int MASK_DEPTH         = 1 << 2;
    protected static final int MASK_CLASS         = 1 << 3;
    protected static final int MASK_RED           = 1 << 4;
    protected static final int MASK_GREEN         = 1 << 5;
    protected static final int MASK_BLUE          = 1 << 6;
    protected static final int MASK_COLORMAP_SIZE = 1 << 7;
    protected static final int MASK_BITS_PER_RGB  = 1 << 8;
    protected static final int MASK_ALL =
	MASK_ID | MASK_SCREEN | MASK_DEPTH | MASK_CLASS |
	MASK_RED | MASK_GREEN | MASK_BLUE |
	MASK_COLORMAP_SIZE | MASK_BITS_PER_RGB;
    private static final int MASK_VISUAL_STRUCTURE = 1 << 31;


    Display display;
    RawData xVisualInfo;
    int infoMask;
    Screen screen;

    Visual(RawData structure, Screen screen, int depth ) {
	this.display = screen.getDisplay();
	this.screen = screen;
	init(structure, depth);
    }

    Visual(Display display, RawData structure, int depth ) {
	this.display = display;
	init(structure, depth);
    }


    protected native void init(RawData structure, int depth);

    protected native void finalize();

    /** This method has package accessibility since the data visual
	structure is only useful for direct Xlib calls.

	@return a pointer to the visual structure.
    */
    native RawData getVisualStructure();
    
    // These methods only make sense if the visual is decomposed.

    /* FIXME: Are these masks encoded in server-, client-, or
       normalized endianess?  */

    public native int getRedMask();
    public native int getGreenMask();
    public native int getBlueMask();

    public native int getScreenNumber();
    public native int getDepth();

    public Screen getScreen() {
	if (screen == null) {
	    screen = new Screen(display, getScreenNumber());
	}
	return screen;
    }
    


    public native int getVisualClass();

    public boolean hasRGBSubfields() {
	switch (getVisualClass()) {
	case VC_TRUE_COLOR:
	case VC_DIRECT_COLOR:
	    return true;
	default:
	    return false;
	}
    }
    

    protected native void ensureXVisualInfo(int requiredMask);

    public String toString() {

	int missingInfo = ~infoMask;
	boolean hasSubfieldInfo =
	    (missingInfo & (MASK_CLASS|MASK_RED|MASK_GREEN|MASK_BLUE)) == 0;

	boolean hasDepth = (missingInfo & MASK_DEPTH) == 0;
	
	return getClass().getName() + "[" +

	    (hasDepth ? "depth=" + getDepth() : "") +

	    (hasRGBSubfields() ?
	     (", redMask=" + Integer.toHexString(getRedMask()) +
	      ", greenMask=" + Integer.toHexString(getGreenMask()) +
	      ", blueMask=" + Integer.toHexString(getBlueMask())) :
	     ", no-subfields") + ", class=" + getVisualClass() +

	    "]";
    }
}
